/*
 ------------------------------------------------------------------------
 idct_ap922sse_opt.c - AP922float iDCT, Intel SSE implementation

 Requirements : 
  Visual C++ 6.0 Pro + Visual Studio Service Pack 4 + Processor Pack Beta
  SSE capable CPU (Pentium-III, Celeron-II)
 ------------------------------------------------------------------------

 AP-922 floating-point iDCT (SSE)
 --------------------------------

  idct_ap922sse_opt.c is an "optimized" version of idct_ap922sse_rawcode.c
  Both files implement functionally the same code.  Persons wanting to
  enhance the performance of AP922sse should work with the raw_code file
  and not *this* file : this "optimized" code is far less readable than 
  the raw_code version.

  For algorithm design notes, please see idct_ap922sse_rawcode.c

 ------------------------------------------
 LEGAL MUMBO-JUMBO : Disclaimer of Warranty
 ------------------------------------------
  This software is available to the user without any license fee or
  royalty on an "as is" basis.  The author disclaims any and all warranties,
  whether express, implied, or statuary, including any implied warranties or 
  merchantability or of fitness for a particular purpose.  In no event shall 
  the copyright-holder be liable for any incidental, punitive, or 
  consequential damages of any kind whatsoever arising from the use of these
  programs.

    This disclaimer of warranty extends to the user of these programs and 
    user's customers, employees, agents, transferees, successors, and assigns.

 Revision history
 ----------------
 09/03/2000 initial release of 'AP922float_SSE' iDCT
            For algorithm design notes, please see idct_ap922sse_rawcode.c

            Despite spending 4 hours hand-scheduling/optimizing SSE assembly,
            Visual C++'s profiler does not report any speed increase.
            DOOHHHH...but anyway here it is.

            IMPORTANT, Visual C++ does not seem to align data to 16-byte
            address offset.  You will need to manually do this!!!
            
 09/03/2000 liaor@iname.com   http://members.tripod.com/~liaor
*/

#pragma warning(disable:4305)

#define NCOS1_16 (0.980785280403230449126182236134239) // cosine(  Pi/16 )
#define NCOS2_16 (0.923879532511286756128183189396788) // cosine( 2Pi/16 )
#define NCOS3_16 (0.831469612302545237078788377617906) // cosine( 3Pi/16 )
#define NCOS4_16 (0.707106781186547524400844362104849) // cosine( 4Pi/16 )
#define NCOS5_16 (0.555570233019602224742830813948533) // cosine( 5Pi/16 )
#define NCOS6_16 (0.382683432365089771728459984030399) // cosine( 6Pi/16 )
#define NCOS7_16 (0.195090322016128267848284868477022) // cosine( 7Pi/16 )

#define TANG1_16 ( NCOS7_16 / NCOS1_16) // tangent( Pi/16)
#define TANG2_16 ( NCOS6_16 / NCOS2_16) // tangent(2Pi/16)
#define TANG3_16 ( NCOS5_16 / NCOS3_16) // tangent(3Pi/16)

// Scaled floating-point coefficient table
//    Original (MMX) iDCT         SSE-iDCT 
//     table-ordering      ->    table-ordering
//    ------------------       ------------------
//    w00, w02, w04, w06,  ->  w00, w01, w02, w03,
//    w01, w03, w05, w07,  ->  w04, w05, w06, w07,
//    w08, w10, w12, w14,  ->  w08, w09, w10, w11,
//    w09, w11, w13, w15,  ->  w12, w13, w14, w15,
//    w16, w18, w20, w22,  ->  w16, w17, w18, w19,
//    w17, w19, w21, w23,  ->  w20, w21, w22, w23,
//    w24, w26, w28, w30,  ->  w24, w25, w26, w27,
//    w25, w27, w29, w31   ->  w28, w29, w30, w31
//
// Compared to the integer-based table, the float table is downscaled by a
// multiplication factor of 0.5
 
#define OVERALL_DOWNSCALE (0.25/(1<<9))
 // all table entries multiplied by an OVERALL_DOWNSCALE of (0.5*0.5)=0.25

#define RS0 (NCOS4_16*OVERALL_DOWNSCALE)  // iDCT row#0 scalar
#define RS1 (NCOS1_16*OVERALL_DOWNSCALE)  // iDCT row#1 scalar
#define RS2 (NCOS2_16*OVERALL_DOWNSCALE)  // iDCT row#2 scalar
#define RS3 (NCOS3_16*OVERALL_DOWNSCALE)  // iDCT row#3 scalar
#define RS4 (NCOS4_16*OVERALL_DOWNSCALE)  // iDCT row#4 scalar
#define RS5 (NCOS3_16*OVERALL_DOWNSCALE)  // iDCT row#5 scalar
#define RS6 (NCOS2_16*OVERALL_DOWNSCALE)  // iDCT row#6 scalar
#define RS7 (NCOS1_16*OVERALL_DOWNSCALE)  // iDCT row#7 scalar

#define DESCALE_SHIFT2 7
//#define DESCALE_SHIFT1 (16+6-DESCALE_SHIFT2)
//#define DESCALE_SHIFT1 (16-9-DESCALE_SHIFT2)

//#define RND_INV_ROWF (1 << (16+6-1) )
#define RND_INV_ROWF (1 << (16-9-1) )
#define RND_FLOAT_OFFSET ((RND_INV_ROWF) - 0.5)

//const static int r_inv_row_sse[2] = {RND_INV_ROWF-1, RND_INV_ROWF-1};
const static int r_inv_row_sse[2] = {0, 0};

const static __int64 useless_padding[1];

const float tab_i_01234567ssef[] = 
{ 
 // Row #0
  NCOS4_16*RS0,  NCOS4_16*RS0,  NCOS4_16*RS0,  NCOS4_16*RS0, 
  NCOS2_16*RS0,  NCOS6_16*RS0, -NCOS6_16*RS0, -NCOS2_16*RS0, 
  NCOS4_16*RS0, -NCOS4_16*RS0, -NCOS4_16*RS0,  NCOS4_16*RS0, 
  NCOS6_16*RS0, -NCOS2_16*RS0,  NCOS2_16*RS0, -NCOS6_16*RS0, 
  NCOS1_16*RS0,  NCOS3_16*RS0,  NCOS5_16*RS0,  NCOS7_16*RS0, 
  NCOS3_16*RS0, -NCOS7_16*RS0, -NCOS1_16*RS0, -NCOS5_16*RS0, 
  NCOS5_16*RS0, -NCOS1_16*RS0,  NCOS7_16*RS0,  NCOS3_16*RS0, 
  NCOS7_16*RS0, -NCOS5_16*RS0,  NCOS3_16*RS0, -NCOS1_16*RS0, 

 // Row #1
  NCOS4_16*RS1,  NCOS4_16*RS1,  NCOS4_16*RS1,  NCOS4_16*RS1, 
  NCOS2_16*RS1,  NCOS6_16*RS1, -NCOS6_16*RS1, -NCOS2_16*RS1, 
  NCOS4_16*RS1, -NCOS4_16*RS1, -NCOS4_16*RS1,  NCOS4_16*RS1, 
  NCOS6_16*RS1, -NCOS2_16*RS1,  NCOS2_16*RS1, -NCOS6_16*RS1, 
  NCOS1_16*RS1,  NCOS3_16*RS1,  NCOS5_16*RS1,  NCOS7_16*RS1, 
  NCOS3_16*RS1, -NCOS7_16*RS1, -NCOS1_16*RS1, -NCOS5_16*RS1, 
  NCOS5_16*RS1, -NCOS1_16*RS1,  NCOS7_16*RS1,  NCOS3_16*RS1, 
  NCOS7_16*RS1, -NCOS5_16*RS1,  NCOS3_16*RS1, -NCOS1_16*RS1, 

 // Row #2
  NCOS4_16*RS2,  NCOS4_16*RS2,  NCOS4_16*RS2,  NCOS4_16*RS2, 
  NCOS2_16*RS2,  NCOS6_16*RS2, -NCOS6_16*RS2, -NCOS2_16*RS2, 
  NCOS4_16*RS2, -NCOS4_16*RS2, -NCOS4_16*RS2,  NCOS4_16*RS2, 
  NCOS6_16*RS2, -NCOS2_16*RS2,  NCOS2_16*RS2, -NCOS6_16*RS2, 
  NCOS1_16*RS2,  NCOS3_16*RS2,  NCOS5_16*RS2,  NCOS7_16*RS2, 
  NCOS3_16*RS2, -NCOS7_16*RS2, -NCOS1_16*RS2, -NCOS5_16*RS2, 
  NCOS5_16*RS2, -NCOS1_16*RS2,  NCOS7_16*RS2,  NCOS3_16*RS2, 
  NCOS7_16*RS2, -NCOS5_16*RS2,  NCOS3_16*RS2, -NCOS1_16*RS2, 

 // Row #3
  NCOS4_16*RS3,  NCOS4_16*RS3,  NCOS4_16*RS3,  NCOS4_16*RS3, 
  NCOS2_16*RS3,  NCOS6_16*RS3, -NCOS6_16*RS3, -NCOS2_16*RS3, 
  NCOS4_16*RS3, -NCOS4_16*RS3, -NCOS4_16*RS3,  NCOS4_16*RS3, 
  NCOS6_16*RS3, -NCOS2_16*RS3,  NCOS2_16*RS3, -NCOS6_16*RS3, 
  NCOS1_16*RS3,  NCOS3_16*RS3,  NCOS5_16*RS3,  NCOS7_16*RS3, 
  NCOS3_16*RS3, -NCOS7_16*RS3, -NCOS1_16*RS3, -NCOS5_16*RS3, 
  NCOS5_16*RS3, -NCOS1_16*RS3,  NCOS7_16*RS3,  NCOS3_16*RS3, 
  NCOS7_16*RS3, -NCOS5_16*RS3,  NCOS3_16*RS3, -NCOS1_16*RS3, 

 // Row #4
  NCOS4_16*RS4,  NCOS4_16*RS4,  NCOS4_16*RS4,  NCOS4_16*RS4, 
  NCOS2_16*RS4,  NCOS6_16*RS4, -NCOS6_16*RS4, -NCOS2_16*RS4, 
  NCOS4_16*RS4, -NCOS4_16*RS4, -NCOS4_16*RS4,  NCOS4_16*RS4, 
  NCOS6_16*RS4, -NCOS2_16*RS4,  NCOS2_16*RS4, -NCOS6_16*RS4, 
  NCOS1_16*RS4,  NCOS3_16*RS4,  NCOS5_16*RS4,  NCOS7_16*RS4, 
  NCOS3_16*RS4, -NCOS7_16*RS4, -NCOS1_16*RS4, -NCOS5_16*RS4, 
  NCOS5_16*RS4, -NCOS1_16*RS4,  NCOS7_16*RS4,  NCOS3_16*RS4, 
  NCOS7_16*RS4, -NCOS5_16*RS4,  NCOS3_16*RS4, -NCOS1_16*RS4, 

 // Row #5
  NCOS4_16*RS5,  NCOS4_16*RS5,  NCOS4_16*RS5,  NCOS4_16*RS5, 
  NCOS2_16*RS5,  NCOS6_16*RS5, -NCOS6_16*RS5, -NCOS2_16*RS5, 
  NCOS4_16*RS5, -NCOS4_16*RS5, -NCOS4_16*RS5,  NCOS4_16*RS5, 
  NCOS6_16*RS5, -NCOS2_16*RS5,  NCOS2_16*RS5, -NCOS6_16*RS5, 
  NCOS1_16*RS5,  NCOS3_16*RS5,  NCOS5_16*RS5,  NCOS7_16*RS5, 
  NCOS3_16*RS5, -NCOS7_16*RS5, -NCOS1_16*RS5, -NCOS5_16*RS5, 
  NCOS5_16*RS5, -NCOS1_16*RS5,  NCOS7_16*RS5,  NCOS3_16*RS5, 
  NCOS7_16*RS5, -NCOS5_16*RS5,  NCOS3_16*RS5, -NCOS1_16*RS5, 

 // Row #6
  NCOS4_16*RS6,  NCOS4_16*RS6,  NCOS4_16*RS6,  NCOS4_16*RS6, 
  NCOS2_16*RS6,  NCOS6_16*RS6, -NCOS6_16*RS6, -NCOS2_16*RS6, 
  NCOS4_16*RS6, -NCOS4_16*RS6, -NCOS4_16*RS6,  NCOS4_16*RS6, 
  NCOS6_16*RS6, -NCOS2_16*RS6,  NCOS2_16*RS6, -NCOS6_16*RS6, 
  NCOS1_16*RS6,  NCOS3_16*RS6,  NCOS5_16*RS6,  NCOS7_16*RS6, 
  NCOS3_16*RS6, -NCOS7_16*RS6, -NCOS1_16*RS6, -NCOS5_16*RS6, 
  NCOS5_16*RS6, -NCOS1_16*RS6,  NCOS7_16*RS6,  NCOS3_16*RS6, 
  NCOS7_16*RS6, -NCOS5_16*RS6,  NCOS3_16*RS6, -NCOS1_16*RS6, 

 // Row #7
  NCOS4_16*RS7,  NCOS4_16*RS7,  NCOS4_16*RS7,  NCOS4_16*RS7, 
  NCOS2_16*RS7,  NCOS6_16*RS7, -NCOS6_16*RS7, -NCOS2_16*RS7, 
  NCOS4_16*RS7, -NCOS4_16*RS7, -NCOS4_16*RS7,  NCOS4_16*RS7, 
  NCOS6_16*RS7, -NCOS2_16*RS7,  NCOS2_16*RS7, -NCOS6_16*RS7, 
  NCOS1_16*RS7,  NCOS3_16*RS7,  NCOS5_16*RS7,  NCOS7_16*RS7, 
  NCOS3_16*RS7, -NCOS7_16*RS7, -NCOS1_16*RS7, -NCOS5_16*RS7, 
  NCOS5_16*RS7, -NCOS1_16*RS7,  NCOS7_16*RS7,  NCOS3_16*RS7, 
  NCOS7_16*RS7, -NCOS5_16*RS7,  NCOS3_16*RS7, -NCOS1_16*RS7, 

 // iDCT column coefficients
  NCOS4_16, NCOS4_16, NCOS4_16, NCOS4_16,
  TANG1_16, TANG1_16, TANG1_16, TANG1_16,
  TANG2_16, TANG2_16, TANG2_16, TANG2_16,
  TANG3_16, TANG3_16, TANG3_16, TANG3_16,
  RND_FLOAT_OFFSET, RND_FLOAT_OFFSET, RND_FLOAT_OFFSET, RND_FLOAT_OFFSET,
  0       , 0       , 0       , 0,        // not used, padding
  0       , 0       , 0       , 0,        // not used, padding
  0       , 0       , 0       , 0         // not used, padding
};

// externally allocated array of 64 floats, for temporary storage
extern void *fTempArray;

void SSE_IDCT( short *data )
{
//  static float scratch[8];
  static const __int64 mmAndmask = 0xFFFF0000FFFF0000; // [ XX __ XX __ ]

  // AP-922 inverse_DCT row transform (floating point version)
  //
  // Row IDCT operator :A_T*M_T*P_T
  // Let Y=[output column data, 8 elements] 32-bit IEEE-754 float
  //     X=[input column data, 8 elements] 16-bit short integer
  //
  //     Y= [ A_T*M_T*P_T ] * X
  //
  //   (Y and X are both column vectors)

#define INPR eax
#define TABLE ecx
#define TABLEP1 esi  // starting address of (TABLE +32 elements)
#define OUTR edx

  __asm {
//    stmxcsr [_mxcsr_backup];  // store MXCSR to mem location _mxcsr_backup

    mov OUTR, dword ptr [fTempArray];
     mov edi, 0x00; // i = 0

//    mov ebx, [_mxcsr_backup];
     sub OUTR, 32;    // precompensate OUT

    mov INPR, dword ptr [data];   ;// row 0
//     and ebx, 0xFFFF9FFF;    // mask out rounding-mode (bits14:13) control

    lea TABLEP1, dword ptr [tab_i_01234567ssef]; // row 0
//     or  ebx, 0x00002000;    // set rounding mode to round-down 14-13 (01b)
    //or  ebx, 0x00000000;    // set rounding mode to nearest 14-13 (01b)

    movq mm7, [mmAndmask];      // mm7 <= [ FFFF 0000 FFFF 0000]
     mov TABLE, TABLEP1;

    add TABLEP1, 128;   // initialize 
//     mov [_mxcsr_fdct], ebx;  // store new mxcsr pattern

//    ldmxcsr [ _mxcsr_fdct ];

  // for ( x = 0; x < 8; ++x )  // transform one row per iteration
//ALIGN 16 ;// align jump address to 16 byte offset
//  acc_idct_colloop1:

//
// begin (0) - Convert short -> float data
//
    movq mm2, [INPR] ;           // (0) mm2 <= x3 x2 x1 x0
     ;//

    movq mm4, [INPR+8] ;         // (0) mm4 <= x7 x6 x5 x4
     pshufw mm0, mm2, 00000000b ;// (0) mm0 <= x0 x0 x0 x0

    pshufw mm1, mm2, 10101010b  ;// (0) mm1 <= x2 x2 x2 x2
     pand mm0, mm7;             ;// (0) mm0 = [ x0 x0 ]*SCALEUP (dword)

    cvtpi2ps xmm0, mm0;         ;// (0) xmm0 <= [__ __ x0 x0] float
// mm0 free
     pand mm1, mm7;             ;// (0) mm1 = [ x2 x2 ]*SCALEUP (dword)

    prefetcht0 [TABLEP1 + 0*32] ;// prefetch coeff +1row, [w7..0]
     pshufw mm0, mm4, 00000000b ;// (0)  mm0 <= x4 x4 x4 x4

    cvtpi2ps xmm1, mm1;         ;// (0) xmm1 <= [__ __ x2 x2] float
// mm1 free
     pand mm0, mm7;             ;// (0) mm4 <= [x4 x4]*SCALEUP (dword)

    prefetcht0 [TABLEP1 + 1*32] ;// prefetch coeff +1row [w15..8]
    movlhps xmm0, xmm0;         ;// (0) xmm0 <= [x0 x0 x0 x0] float
     cvtpi2ps xmm2, mm0;        ;// (0) xmm2 <= [__ __ x4 x4] float
// mm0 free

// 1) Apply M_T*P_T operator (P_T is implicit with table/element order)
//    (table order has been changed for optimum SSE execution)

    mulps xmm0, [TABLE + 0*16]  ;// (1a) xmm0 <= a_v0 = [w3..w0] * x0
     pshufw mm1, mm4, 10101010b; // (0) mm1 <= [x6 x6 x6 x6]

    pand mm1, mm7;              ;// (0) mm1 <= [x6 x6]*SCALEUP (dword)
     movlhps xmm1, xmm1;        ;// (0) xmm1 <=[ x2 x2 x2 x2] float

    mulps xmm1, [TABLE + 1*16]  ;// (1a) xmm1 <= a_v1 = [w7..w4] * x2
     cvtpi2ps xmm3, mm1;        ;// (0) xmm3 <=[ __ __ x6 x6] float
// mm1 free

    movlhps xmm2, xmm2;         ;// (0) xmm2 <= [x4 x4 x4 x4] float
     pshufw mm0, mm2, 01010101b; // (0) mm1 <=[x1 x1 x1 x1]

    mulps xmm2, [TABLE + 2*16];  // (1a) xmm2 <= a_v2 = [w11..w8] * x4
     pshufw mm1, mm2, 11111111b; // (0) mm1 <=[x3 x3 x3 x3]
// mm2 free

    pand mm0, mm7;               // (0) mm0 <=[x1 x1]*SCALEUP (dword)
     movlhps xmm3, xmm3;        ;// (0) xmm3 <= [x6 x6 x6 x6]

    mulps xmm3, [TABLE + 3*16] ; // (1a) xmm3 <= a_v3 = [w15..w12] * x6
     cvtpi2ps xmm4, mm0;         // (0) xmm4 <=[__ __ x1 x1] float
// mm0 free

    /////////////////////////////////////////////
    //
    //  loop point 
    //  for ( i = 0; i < 8; i=i+1 )  
    //  {
    //     ....// idct_row(#i)

ALIGN 16 ;// align jump address to 16 byte offset
  acc_idct_rowloop1:

    addps xmm1, xmm0;           ;// (1b) xmm1 <= a_v0 + a_v1
// xmm0 free
     pand mm1, mm7;              // (0) mm1 <=[x3 x3]*SCALEUP (dword)

     pshufw mm0, mm4, 01010101b; // (0) mm0 <= [x5 x5 x5 x5]

    cvtpi2ps xmm5, mm1;          // (0) xmm5 <=[__ __ x3 x3] float
     pand mm0, mm7;              // (0) mm0 <= [x5 x5]*SCALEUP (dword)
// mm1 free

    cvtpi2ps xmm6, mm0;          // (0) xmm6 <=[__ __ x5 x5] float
 // mm0 free
     movlhps xmm4, xmm4;         // (0) xmm4 <= [x1 x1 x1 x1]

    mulps xmm4, [TABLE + 4*16]  ;// (1a) xmm4 <= b_v0
     pshufw mm1, mm4, 11111111b; // (0) mm1 <= [x7 x7 x7 x7]
// mm4 free

    addps xmm2, xmm1;           ;// (1b) xmm2 + a_v2 <= (a_v0+a_v1) + a_v2
// xmm1 free
     movlhps xmm5, xmm5;        ;// (0) xmm5 <=[x3 x3 x3 x3] float

    mulps xmm5, [TABLE + 5*16]  ;// (1a) xmm5 <= b_v1
     pand mm1, mm7;             ;// (0) mm1 <= [x7 x7]*SCALEUP (dword)

    cvtpi2ps xmm7, mm1;         ;// (0) xmm7 <= [__ __ x7 x7]
// mm1 free
     movlhps xmm6, xmm6;         // (0) xmm6 <= [x5 x5 x5 x5] float

    mulps xmm6, [TABLE + 6*16]  ;// (1a) xmm6 <= b_v2
     add OUTR, 32;              ;// OUT <= OUT + 32 (increment row) 

    addps xmm3, xmm2;           ;// (1b) xmm3 + a_v3 <=(a_v0+a_v1+a_v2) + a_v3
// xmm3 = complete a-vector ( a_v[3..0] )
// xmm2 free
     add INPR, 16;               // increment INP(psrc) + 1row 

    prefetcht0 [TABLEP1 + 2*32] ;// prefetch coeff +1row [w23..15]
     movlhps xmm7, xmm7;

    mulps xmm7, [TABLE + 7*16]  ;// (1a) xmm7 <= b_v3
     addps xmm5, xmm4;          ;// (1b) xmm5 <= b_v0 + b_v1
// xmm4 free

    ///////////////////////////////////////////////////////////////
    //
    // begin processing next row ("+1")
    //
    // the following code intermixes processing from row (x) and (x+1)
    // This is the result of software pipelining to reduce wasted
    // cycles.

    movq mm2, [INPR] ;           // x+1 (0) mm2 <= x3 x2 x1 x0
     add TABLE, 128;              // move TABLE to w[32]

    prefetcht0 [TABLEP1 + 3*32] ;// prefetch coeff +1row [w31..24]
     add TABLEP1, 128;           // move TABLEP1 to w[32]

// xmm4 used
    movaps xmm4, xmm3;           // (1b) xmm4 <= a_v[3..0] 
     pshufw mm0, mm2, 00000000b ;// x+1 (0) mm0 <= x0 x0 x0 x0

    movq mm4, [INPR+8] ;         // x+1 (0) mm4 <= x7 x6 x5 x4
     addps xmm6, xmm5;           // (1b) xmm6 + b_v2 <= (b_v0+b_v1) + b_v2
// xmm5 free

    pshufw mm1, mm2, 10101010b  ;// x+1 (0) mm1 <= x2 x2 x2 x2
     pand mm0, mm7;             ;// x+1 (0) mm0 = [ x0 x0 ]*SCALEUP (dword)

    cvtpi2ps xmm0, mm0;         ;// x+1 (0) xmm0 <= [__ __ x0 x0] float
// mm0 free
     pand mm1, mm7;             ;// x+1 (0) mm1 = [ x2 x2 ]*SCALEUP (dword)

// xmm5 used
     movaps xmm5, xmm3;         ;// xmm5 <= a_v[3..0]
// xmm3 free
     add edi, 1;                 // i = i + 1 (loop index)

    prefetcht0 [TABLEP1 + 0*32] ;// x+1 prefetch coeff +1row, [w7..0]
     pshufw mm0, mm4, 00000000b ;// x+1 (0)  mm0 <= x4 x4 x4 x4

    addps xmm7, xmm6;            // (1b) xmm7 + b_v3 <=(b_v0+b_v1+b_v2) + b_v3
// xmm6 free
// xmm7 = complete b-vector ( b_v[3..0] )
     pand mm0, mm7;             ;// x+1 (0) mm4 <= [x4 x4]*SCALEUP (dword)

    prefetcht0 [TABLEP1 + 1*32] ;// x+1 prefetch coeff +1row [w15..8]
     movlhps xmm0, xmm0;        ;// x+1 (0) xmm0 <= [x0 x0 x0 x0] float

// 1) Apply M_T*P_T operator (P_T is implicit with table/element order)
//    (table order has been changed for optimum SSE execution)

    mulps xmm0, [TABLE + 0*16]  ;// x+1 (1a) xmm0 <= a_v0 = [w3..w0] * x0
     cvtpi2ps xmm1, mm1;        ;// x+1 (0) xmm1 <= [__ __ x2 x2] float
// mm1 free

    cvtpi2ps xmm2, mm0;         ;// x+1 (0) xmm2 <= [__ __ x4 x4] float
// mm0 free
     ;//

// output y[4..7] = a[7..4] + b[7..4], output order must be reversed!
    subps xmm5, xmm7;       ;// (2a) xmm5 <= a_v[3..0]-b_v[3..0] (y4 y5 y6 y7)
     ;//

    movlhps xmm1, xmm1;         ;// x+1 (0) xmm1 <=[ x2 x2 x2 x2] float
     pshufw mm1, mm4, 10101010b ;// x+1 (0) mm1 <= [x6 x6 x6 x6]
    
    mulps xmm1, [TABLE + 1*16]  ;// x+1 (1a) xmm1 <= a_v1 = [w7..w4] * x2
// output y[3..0] = a[3..0] + b[3..0]
     addps xmm4, xmm7;      ;// (2a) xmm4 <= a_v[3..0]+b_v[3..0] (y3 y2 y1 y0)
// xmm7 free

    movlhps xmm2, xmm2;         ;// x+1 (0) xmm2 <= [x4 x4 x4 x4] float
     pand mm1, mm7;              ;// x+1 (0) mm1 <= [x6 x6]*SCALEUP (dword)

    mulps xmm2, [TABLE + 2*16];  // x+1 (1a) xmm2 <= a_v2 = [w11..w8] * x4
// fix output order, y[4..7] -> y[7..4]
     shufps xmm5, xmm5, 00011011b; // (2a) xmm5 [y7 y6 y5 y4] <- [y4 y5 y6 y7]

    cvtpi2ps xmm3, mm1;           // x+1 (0) xmm3 <=[ __ __ x6 x6] float
// mm1 free
     ;//

// !*@&#@ Visual C++ refuses to align data to 16-byte offset, SIGH

    pshufw mm0, mm2, 01010101b;   // x+1 (0) mm1 <=[x1 x1 x1 x1]
    movaps [OUTR + 0*16], xmm4;   // (2b) store [y3 y2 y1 y0]
//     movups [OUTR + 0*16], xmm4;   // (2b) store [y3 y2 y1 y0]
// xmm4 free

    pshufw mm1, mm2, 11111111b;   // x+1 (0) mm1 <=[x3 x3 x3 x3]
// mm2 free
     pand mm0, mm7;               // x+1 (0) mm0 <=[x1 x1]*SCALEUP (dword)

    movlhps xmm3, xmm3;          ;// x+1 (0) xmm3 <= [x6 x6 x6 x6]
     cvtpi2ps xmm4, mm0;          // x+1 (0) xmm4 <=[__ __ x1 x1] float
// mm0 free

    mulps xmm3, [TABLE + 3*16]   ;// x+1 (1a) xmm3 <= a_v3 = [w15..w12] * x6
    movaps [OUTR + 1*16], xmm5;  ;// (2b) store [y7 y6 y5 y4
//     movups [OUTR + 1*16], xmm5;  ;// (2b) store [y7 y6 y5 y4
// xmm5 free
    cmp edi, 0x07;
       // end for ( x = 0; x < 7; ++x )  // transform one row per iteration
    jl  acc_idct_rowloop1;

// Since the above code integrates processing for rows(x) and (x+1),
// the last row(#7) must be "unrolled" without the (x+1) code.

    //
    //
    /////////////////////////////////////////////

    /////////////////////////////////////////////
    //
    //  iDCT row #7

    addps xmm1, xmm0;           ;// (1b) xmm1 <= a_v0 + a_v1
// xmm0 free
     pand mm1, mm7;              // (0) mm1 <=[x3 x3]*SCALEUP (dword)

// this prefetch retrieves the TAN1, TAN2, TAN3, COS4 coefficients
// from the upcoming col_idct() operation
    prefetcht0 [TABLEP1 + 0*32] ;// x+1 prefetch coeff COL COEFFICIENTS
     pshufw mm0, mm4, 01010101b; // (0) mm0 <= [x5 x5 x5 x5]

    cvtpi2ps xmm5, mm1;          // (0) xmm5 <=[__ __ x3 x3] float
     pand mm0, mm7;              // (0) mm0 <= [x5 x5]*SCALEUP (dword)
// mm1 free

    cvtpi2ps xmm6, mm0;          // (0) xmm6 <=[__ __ x5 x5] float
 // mm0 free
     movlhps xmm4, xmm4;         // (0) xmm4 <= [x1 x1 x1 x1]

    mulps xmm4, [TABLE + 4*16]  ;// (1a) xmm4 <= b_v0
     pshufw mm1, mm4, 11111111b; // (0) mm1 <= [x7 x7 x7 x7]
// mm4 free

    addps xmm2, xmm1;           ;// (1b) xmm2 + a_v2 <= (a_v0+a_v1) + a_v2
// xmm1 free
     movlhps xmm5, xmm5;        ;// (0) xmm5 <=[x3 x3 x3 x3] float

    mulps xmm5, [TABLE + 5*16]  ;// (1a) xmm5 <= b_v1
     pand mm1, mm7;             ;// (0) mm1 <= [x7 x7]*SCALEUP (dword)

    cvtpi2ps xmm7, mm1;         ;// (0) xmm7 <= [__ __ x7 x7]
// mm1 free
     movlhps xmm6, xmm6;         // (0) xmm6 <= [x5 x5 x5 x5] float

    mulps xmm6, [TABLE + 6*16]  ;// (1a) xmm6 <= b_v2
     add OUTR, 32;              ;// OUT <= OUT + 32 (increment row) 

    addps xmm3, xmm2;           ;// (1b) xmm3 + a_v3 <=(a_v0+a_v1+a_v2) + a_v3
// xmm3 = complete a-vector ( a_v[3..0] )
// xmm2 free

// this prefetch retrieves the TAN1, TAN2, TAN3, COS4 coefficients
// from the upcoming col_idct() operation
    prefetcht0 [TABLEP1 + 1*32] ;// prefetch COL coefficients
     movlhps xmm7, xmm7;
// TABLEP1 (esi) free

    mulps xmm7, [TABLE + 7*16]  ;// (1a) xmm7 <= b_v3
     addps xmm5, xmm4;          ;// (1b) xmm5 <= b_v0 + b_v1
// xmm4 free

;/////////////////////////////////////////////////////////////////////
;//
;//
;// Begin iDCT column transform (columns 0-3)
;//
;//
;/////////////////////////////////////////////////////////////////////
    add TABLE, 128; // COL SETUP
     mov edi, 0x00; // i = 0; // COL SETUP

  //  TABLE now points to the proper location, no need to reload with lea
  //  lea TABLE, [tab_i_01234567ssef];; // COL SETUP
 //   add TABLE, 1024; ; // COL SETUP// increment TABLE to the COS4 entry

#define SCRATCH esi  // column iDCT
#define INPC eax     // column iDCT
#define OUTC edx     // column iDCT
#define _PCOS4        TABLE   
#define _PTAN1        TABLE+16
#define _PTAN2        TABLE+32
#define _PTAN3        TABLE+48
#define _PRND_INV_ROW TABLE+64

    mov INPC, dword ptr [fTempArray]; ; // COL SETUP
     addps xmm6, xmm5;           // (1b) xmm6 + b_v2 <= (b_v0+b_v1) + b_v2
// xmm5 free

//    lea SCRATCH, [scratch]    ; // COL SETUP
    mov SCRATCH, INPC;           ;// SCRATCH PAD writes over temp data
// xmm4 used
     movaps xmm4, xmm3;           // (1b) xmm4 <= a_v[3..0] 

    movaps xmm0, [INPC + 1*32] ; // col (1) xmm0 <= dp[4]
// xmm5 used
     movaps xmm5, xmm3;         ;// xmm5 <= a_v[3..0]
// xmm3 free

    addps xmm7, xmm6;            // (1b) xmm7 + b_v3 <=(b_v0+b_v1+b_v2) + b_v3
// xmm6 free
// xmm7 = complete b-vector ( b_v[3..0] )
     ;//

// output y[4..7] = a[7..4] + b[7..4], output order must be reversed!
    movaps xmm2, [INPC + 3*32];  // col (1) xmm2 <= dp[6]
     ;

    ;

    movaps xmm3, [INPC + 5*32] ; // col (1) xmm3 <= dp[7]
     ;

    addps xmm4, xmm7;      ;// (2a) xmm4 <= a_v[3..0]+b_v[3..0] (y3 y2 y1 y0)
     ;

    ;// COLUMN IDCT
    ;// 1) Apply (D_T * P_T) - the cos() coefficients of D_T are implicit
    ;//    in the idct_row operation.  But we still need to apply the
    ;//    shuffling operation of D_T.

    ;
     ;

// output y[3..0] = a[3..0] + b[3..0]
    subps xmm5, xmm7;       ;// (2a) xmm5 <= a_v[3..0]-b_v[3..0] (y4 y5 y6 y7)
// xmm7 free

// fix output order, y[4..7] -> y[7..4]
    shufps xmm5, xmm5, 00011011b; // (2a) xmm5 [y7 y6 y5 y4] <- [y4 y5 y6 y7]
     ;

    movaps xmm6, xmm2;         ;// col (2a) xmm6 <= dp[6]
  // movaps xmm1, [INPC + 7*32];  // col (1) xmm1 <= dp[5]
     movaps xmm1, xmm4;            // col (1) xmm1 <= dp[5]

// !*@&#@ Visual C++ refuses to align data to 16-byte offset, SIGH

    movaps [OUTR + 0*16], xmm4;  // (2b) store [y3 y2 y1 y0]
     movaps xmm4, xmm0;          // col (1) xmm4 <= dp[4]
//     movups [OUTR + 0*16], xmm4;   // (2b) store [y3 y2 y1 y0]
// xmm4 free
// xmm4 used 
    

    // 2) Apply B_T
    //
    //    b[4] =   dp[4] + ( dp[5]*tg_1_16f );
    //    b[5] = - dp[5] + ( dp[4]*tg_1_16f );
    //    b[6] =   dp[6] + ( dp[7]*tg_3_16f );
    //    b[7] = - dp[7] + ( dp[6]*tg_3_16f );

    mulps xmm4, [_PTAN1];    ;// (2a) xmm4 <= (dp[4]*tg_1_16f)
     ;

    movaps [OUTR + 1*16], xmm5; ;// (2b) store [y7 y6 y5 y4]
//     movups [OUTR + 1*16], xmm5;  ;// (2b) store [y7 y6 y5 y4]
// xmm5 free
// xmm5 used
     movaps xmm5, xmm1       ;// (2a) xmm5 <= dp[5] 

   ////////////////////////////////////
   // end of iDCT row #7 processing
   ////////////////////////////////////

    mulps xmm5, [_PTAN1];    ;// (2a) xmm5 <= (dp[5]*tg_1_16f)

    ;

    mulps xmm6, [_PTAN3];    ;// (2a) xmm6 <= (dp[6]*tg_3_16f)
     movaps xmm7, xmm3;      ;// (2a) xmm7 <= dp[7]

    subps xmm4, xmm1;        ;// (2b) xmm4 <= b[5] = (dp[4]*tg_1_16f) - dp[5]
// xmm1 free
     ;

     
    mulps xmm7, [_PTAN3];    ;// (2a) xmm7 <= (dp[7]*tg_3_16f)
     ;

    addps xmm5, xmm0;        ;// (2b) xmm5 <= b[4] = (dp[5]*tg_1_16f) + dp[4]
// xmm0 free

    ;


    mov OUTC, dword ptr [data];   ;//; // COL SETUP row 0
    subps xmm6, xmm3;        ;// (2b) xmm6 <= b[7] = (dp[6]*tg_1_16f) - dp[7]
// xmm3 free


// xmm1 used
     movaps xmm1, xmm4;      ;// (2b) xmm1 <= b[5]

// xmm3 used
    movaps xmm3, [INPC + 6*32]   ;// (2a) xmm3 <= dp[3]
     ;

    addps xmm7, xmm2;         // (2b) xmm7 <= b[6] = (dp[7]*tg_1_16f) + dp[6]
// xmm2 free
// xmm0 used
     movaps xmm0, xmm5;      ;// (2b) xmm0 <= b[4]

    ;

    // 3) Apply E_T
    //
    //     e[4] = b[4] + b[6];
    //     e[5] = b[4] - b[6];
    //     e[6] = b[5] + b[7];
    //     e[7] = b[5] - b[7];

// xmm2 used
    movaps xmm2, xmm3;          ;// (2a) xmm2 <= dp[3]
     addps xmm1, xmm6;          ;// (3) xmm1 <= e[6] = b[5] + b[7]

    mulps xmm3, [_PTAN2];       ;// (2a) xmm3 <= (dp[3]*tg_2_16f)
     ;

    subps xmm5, xmm7;           ;// (3) xmm5 <= e[5] = b[4] - b[6]

    ;

    subps xmm4, xmm6;           ;// (3) xmm4 <= e[7] = b[5] - b[7]
// xmm6 free
// xmm6 used
     movaps xmm6, [INPC + 2*32] ;// (2a) xmm6 <= dp[2]

    ;

    addps xmm0, xmm7;           ;// (3) xmm0 <= e[4] = b[4] + b[6]
// xmm7 free
// xmm7 used
     movaps xmm7, xmm5;         ;// (3) xmm7 <= e[5]

    // 4) Apply F_T
    //
    //    f[4] = e[4]
    //    f[5] = (e[5] + e[6]) * cos_4_16f;
    //    f[6] = (e[5] - e[6]) * cos_4_16f;
    //    f[7] = e[7]

// xmm0 <= f[4] = e[4]
// xmm4 <= f[7] = e[7]

    ;

    addps xmm5, xmm1;           ;// (4a) xmm5 <= (e[5]+e[6])

    ;

    subps xmm7, xmm1;           ;// (4a) xmm7 <= (e[5]-e[6])
// xmm1 free
// xmm1 used
     movaps xmm1, [INPC + 0*32] ;// (1) xmm1 <= dp[0]

    movaps [SCRATCH], xmm0;         ;// store f[4] in scratch space
// xmm0 free (in scratch)

    mulps xmm5, [_PCOS4]     ;// (4b) xmm5 <= f[5] = (e[5] + e[6]) * cos_4_16f;
     addps xmm3, xmm6;       ;// (2b) xmm3 <= b[2] = (dp[3]*tg_2_16f) + dp[2]

//movups [SCRATCH+3*16], xmm4;  ;// store f[7] in scratch space
// xmm4 f[7] free (in scratch)

    ;

    mulps xmm6, [_PTAN2];       ;// (2a) xmm6 <= (dp[2]*tg_2_16f)

    ;

    mulps xmm7, [_PCOS4]        ;// (4b) xmm7 <= f[6] = (e[5] - e[6]) * cos_4_16f;
    
    ;

    subps xmm6, xmm2;           ;// (2b) xmm6 <= b[3] = (dp[2]*tg_2_16f) - dp[3]
// xmm2 free
// xmm2 used
//movups [SCRATCH+1*16], xmm5;  ;// xmm5 <= store f[5] in scratch space
// xmm5 f[5] free (in scratch)
     movaps xmm2, [INPC+4*32];  ;// (1) xmm2 <= dp[1]

// xmm0 used
    movaps xmm0, xmm1;          ;// (1) xmm0 <= dp[0]
     ;

    addps xmm1, xmm2;           ;// (2b) xmm1 <= b[0] = dp[0] + dp[1]

// movups [SCRATCH+2*16], xmm7; ;// xmm7 <= store f[6] in scratch space
// xmm7 f[6] free (in scratch)

    ;

//    movups [INP+1*16], xmm1;  ;// temp store

    subps xmm0, xmm2;           ;// (2b) xmm0 <= b[1] = dp[0] - dp[1]
// xmm2 free

//    e[0] = b[0] + b[2];
//    e[1] = b[1] + b[3];
//    e[2] = b[1] - b[3];
//    e[3] = b[0] - b[2];

    ;

// xmm2 used 
    movaps xmm2, xmm1;          ;// (2b) xmm1 <= b[0]
     addps xmm1, xmm3;          ;// (3) xmm1 <= e[0] = b[0] + b[2];

    ;

    subps xmm2, xmm3;           ;// (3) xmm3 <= e[3] = b[0] - b[2];
// xmm3 free
// xmm3 used
     movaps xmm3, xmm0;         ;// (2b) xmm3 <= b[1]

    ;

    addps xmm0, xmm6;           ;// (3) xmm0 <= e[1] = b[1] + b[3];

    ;

    subps xmm3, xmm6;           ;// (3) xmm3 <= e[2] = b[1] - b[3];
// xmm6 free
// xmm6 used
     movaps xmm6, xmm2;         ;// (4) xmm6 <= f[3]

// xmm1 <= f[0]
// xmm0 <= f[1]
// xmm3 <= f[2]
// xmm2 <= f[3]
/*

    y[0*8] = SHIFT_ROUND_COLF( _F0 + _F4 );
    y[1*8] = SHIFT_ROUND_COLF( _F1 + _F5 );
    y[2*8] = SHIFT_ROUND_COLF( _F2 + _F6 );
    y[3*8] = SHIFT_ROUND_COLF( _F3 + _F7 );

    y[4*8] = SHIFT_ROUND_COLF( _F3 - _F7 );
    y[5*8] = SHIFT_ROUND_COLF( _F2 - _F6 );
    y[6*8] = SHIFT_ROUND_COLF( _F1 - _F5 );
    y[7*8] = SHIFT_ROUND_COLF( _F0 - _F4 );

    6a) round compensation, adjust float value to correct round alignment
        addps (ROUND_COMP - 0.5)

    6b) float -> int32 conversion (cvtps2pi)
       cvtps2pi mm0, xmm0;  // mm0 <= int32 versions of y[1 0]
        movhlps xmm0, xmm0; 
       cvtps2pi mm1, xmm0;  // mm1 <= int32 versions of y[3 2]

    6c) int32 -> int16 conversion (packssdw)
        packssdw mm0, mm1; // mm0 <= int16 versions of y[3 2 1 0]

    6d) int16 decsale (psraw)
       psraw mm0, DESCALE_SHIFT2; // mm0 <= final descaled output, {-256,+255}

    ** The descale shift is chosen to implement a {-256,+255} output range.

*/
    ;

    addps xmm2, xmm4;           ;// (5) xmm2 <= y[3] = f[3] + f[7]

    ;

    subps xmm6, xmm4;           ;// (5) xmm6 <= y[4] = f[3] - f[7]
// xmm4 free
// xmm4 used
     movaps xmm4, xmm0;         ;// xmm0 <= f[1]

    ;

    addps xmm2, [_PRND_INV_ROW]; // (6a) ROUND compensation

    ;

    addps xmm0, xmm5;           ;// (5) xmm0 <= y[1] = f[1] + f[5]

    ;

   cvtps2pi mm0, xmm2;          ;// (6b) mm0 <= y[3] low
    movhlps xmm2, xmm2;

   cvtps2pi mm1, xmm2;          ;// (6b) mm1 <= y[3] high
    ;

// xmm2 free
   addps xmm0, [_PRND_INV_ROW]; ;// (6a) ROUND compensation

   ;

   subps xmm4, xmm5;            ;// (5) xmm4 <= y[6] = f[1] - f[5]
// xmm5 free
// xmm2 used
    movaps xmm2, [SCRATCH];     ;// xmm2 <= f[4]

   packssdw mm0, mm1;           ;// (6c) mm0 <= y[3] 
    ;

   cvtps2pi mm2, xmm0;          ;// (6b) mm2 <= y[1] low
    movhlps xmm0, xmm0;

   cvtps2pi mm3, xmm0;          ;// (6b) mm3 <= y[1] high
// xmm0 free
// xmm0 used
// xmm5 used
    movaps xmm0, [_PRND_INV_ROW];

   movaps xmm5, xmm3;          ;// xmm5 <= f[2]
    addps xmm3, xmm7;            ;// (5) xmm3 <= y[2] = f[2] + f[6]

   psraw mm0, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    add INPC, 16;

   addps xmm6, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    ;

   packssdw mm2, mm3;           ;// (6c) mm2 <= y[1]
    ;

   subps xmm5, xmm7;            ;// (5) xmm5 <= y[5] = f[2] - f[6]
// xmm7 free
    movq [OUTC + 3*16], mm0;    ;// mm0 <= y[3] stored

   psraw mm2, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   cvtps2pi mm4, xmm6;          ;// (6b) mm4 <= y[4] low
    movhlps xmm6, xmm6;

   cvtps2pi mm5, xmm6;          ;// (6b) mm5 <= y[4] high
// xmm6 free

   ////////////////////////////////////
   // Start of iDCT column 4-7 processing (denote "col+4")
   ////////////////////////////////////

// xmm6 used
    movaps xmm6, [INPC + 3*32];  // col+4 (1) xmm6 <= dp[6]
  
   addps xmm3, xmm0             ;// (6a) ROUND compensation
    movq [OUTC + 1*16], mm2;    ;// y[1] stored

// xmm7 used
   movaps xmm7, xmm1;           ;// xmm7 <= f[0]
    ;

   addps xmm1, xmm2;            ;// (5) xmm2 <= y[0] = f[0] + f[4]
    ;

   packssdw mm4, mm5;           ;// (6c) mm4 <= y[4]
    ;

   cvtps2pi mm6, xmm3;          ;// (6b) mm6 <= y[2] low
    movhlps xmm3, xmm3;

   cvtps2pi mm7, xmm3;          ;// (6b) mm7 <= y[2] high
// xmm3 free
// xmm3 used
    movaps xmm3, [INPC + 5*32] ; // col+4 (1) xmm3 <= dp[7]

   subps xmm7, xmm2;            ;// (5) xmm7 <= y[7] = f[0] - f[4]
// xmm2 free
// xmm2 used
     movaps xmm2, xmm6;         ;// col+4 (2a) xmm2 <= dp[6]

    // 2) Apply B_T (for columns 4-7)
    //
    //    b[4] =   dp[4] + ( dp[5]*tg_1_16f );
    //    b[5] = - dp[5] + ( dp[4]*tg_1_16f );
    //    b[6] =   dp[6] + ( dp[7]*tg_3_16f );
    //    b[7] = - dp[7] + ( dp[6]*tg_3_16f );

   psraw mm4, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    mulps xmm6, [_PTAN3];       ;// col+4 (2a) xmm6 <= (dp[6]*tg_3_16f)

   addps xmm5, xmm0;            ;// (6a) ROUND compensation
    movq [OUTC + 4*16], mm4;    ;// y[4] stored

   packssdw mm6, mm7;           ;// (6c) mm6 <= y[2]
    ;

   addps xmm4, xmm0             ;// (6a) ROUND compensation
    ;

   psraw mm6, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   cvtps2pi mm0, xmm5;          ;// (6b) mm0 <= y[5] low
    movhlps xmm5, xmm5;

   cvtps2pi mm1, xmm5;          ;// (6b) mm1 <= y[5] high
// xmm5 free
// xmm5 used
   movaps xmm5, [INPC + 7*32];  // col+4 (1) xmm5 <= dp[5]

   cvtps2pi mm2, xmm4;          ;// (6b) mm2 <= y[6] low
    movhlps xmm4, xmm4;

   cvtps2pi mm3, xmm4;          ;// (6b) mm3 <= y[6] high
// xmm4 free
// xmm4 used
    movaps xmm4, [INPC + 1*32] ; // col+4 (1) xmm4 <= dp[4]

   movq [OUTC + 2*16], mm6;    ;// y[2] stored
    addps xmm1, xmm0             ;// (6a) ROUND compensation

   packssdw mm0, mm1;           ;// (6c) mm0 <= y[5]
    ;

   addps xmm7, xmm0             ;// (6a) ROUND compensation
// xmm0 free
// xmm0 used
    movaps xmm0, xmm4;          // col+4 (1) xmm0 <= dp[4]

   psraw mm0, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    mulps xmm4, [_PTAN1];    ;// col+4 (2a) xmm4 <= (dp[4]*tg_1_16f)

   cvtps2pi mm4, xmm1;          ;// (6b) mm4 <= y[0] low
    movhlps xmm1, xmm1;

   packssdw mm2, mm3;           ;// (6c) mm2 <= y[6]
    ;

   cvtps2pi mm5, xmm1;          ;// (6b) mm5 <= y[0] high
// xmm1 free
    movq [OUTC + 5*16], mm0;

// xmm1 used
   movaps xmm1, xmm5           ;// col+4 (2a) xmm1 <= dp[5] 
    psraw mm2, DESCALE_SHIFT2;   // (6d) clip y[] to {-256,+255}

   cvtps2pi mm6, xmm7;         ;// (6b) mm6 <= y[7] low
    movhlps xmm7, xmm7;

   cvtps2pi mm7, xmm7;         ;// (6b) mm7 <= y[7] high
// xmm7 free
    mulps xmm5, [_PTAN1];    ;// col+4 (2a) xmm5 <= (dp[5]*tg_1_16f)

   subps xmm4, xmm1;   ;// col+4 (2b) xmm4 <= b[5] = (dp[4]*tg_1_16f) - dp[5]
// xmm1 free
    movaps xmm7, xmm3; ;// col+4 (2a) xmm7 <= dp[7]

   packssdw mm4, mm5;          ;// (6c) mm4 <= y[0]
    mulps xmm7, [_PTAN3];    ;// col+4 (2a) xmm7 <= (dp[7]*tg_3_16f)

   subps xmm6, xmm3;    ;// col+4 (2b) xmm6 <= b[7] = (dp[6]*tg_1_16f) - dp[7]
// xmm3 free
    movq [OUTC + 6*16], mm2;
 
// xmm3 used
   movaps xmm3, [INPC + 6*32]   ;// col+4 (2a) xmm3 <= dp[3]
    psraw mm4, DESCALE_SHIFT2;   // (6d) clip y[] to {-256,+255}

// xmm1 used
   movaps xmm1, xmm4;   ;// col+4 (2b) xmm1 <= b[5]
    addps xmm5, xmm0;  ;// col+4 (2b) xmm5 <= b[4] = (dp[5]*tg_1_16f) + dp[4]

   movq [OUTC + 0*16], mm4;
    packssdw mm6, mm7;          ;// (6c) mm6 <= y[7]

   psraw mm6, DESCALE_SHIFT2;   // (6d) clip y[] to {-256,+255}
    addps xmm7, xmm2;   // col+4 (2b) xmm7 <= b[6] = (dp[7]*tg_1_16f) + dp[6]
// xmm2 free

// xmm0 used
   movaps xmm0, xmm5;      ;// col+4 (2b) xmm0 <= b[4]

 
    // 3) Apply E_T
    //
    //     e[4] = b[4] + b[6];
    //     e[5] = b[4] - b[6];
    //     e[6] = b[5] + b[7];
    //     e[7] = b[5] - b[7];

// xmm2 used
     movaps xmm2, xmm3;          ;// col+4 (2a) xmm2 <= dp[3]

    addps xmm1, xmm6;          ;// col+4 (3) xmm1 <= e[6] = b[5] + b[7]
     movq [OUTC + 7*16], mm6;

   ////////////////////////////////////
   // end of iDCT column 0-3 processing
   ////////////////////////////////////

    mulps xmm3, [_PTAN2];       ;// (2a) xmm3 <= (dp[3]*tg_2_16f)
     add OUTC, 8;               ;// advance output to columns 4-7

    subps xmm5, xmm7;           ;// (3) xmm5 <= e[5] = b[4] - b[6]

    ;

    subps xmm4, xmm6;           ;// (3) xmm4 <= e[7] = b[5] - b[7]
// xmm6 free
// xmm6 used
     movaps xmm6, [INPC + 2*32] ;// (2a) xmm6 <= dp[2]

    ;

    addps xmm0, xmm7;           ;// (3) xmm0 <= e[4] = b[4] + b[6]
// xmm7 free
// xmm7 used
     movaps xmm7, xmm5;         ;// (3) xmm7 <= e[5]

    // 4) Apply F_T
    //
    //    f[4] = e[4]
    //    f[5] = (e[5] + e[6]) * cos_4_16f;
    //    f[6] = (e[5] - e[6]) * cos_4_16f;
    //    f[7] = e[7]

// xmm0 <= f[4] = e[4]
// xmm4 <= f[7] = e[7]

    ;

    addps xmm5, xmm1;           ;// (4a) xmm5 <= (e[5]+e[6])

    ;

    subps xmm7, xmm1;           ;// (4a) xmm7 <= (e[5]-e[6])
// xmm1 free
// xmm1 used
     movaps xmm1, [INPC + 0*32] ;// (1) xmm1 <= dp[0]

    movaps [SCRATCH], xmm0;         ;// store f[4] in scratch space
// xmm0 free (in scratch)

    mulps xmm5, [_PCOS4]     ;// (4b) xmm5 <= f[5] = (e[5] + e[6]) * cos_4_16f;
     addps xmm3, xmm6;       ;// (2b) xmm3 <= b[2] = (dp[3]*tg_2_16f) + dp[2]

//movups [SCRATCH+3*16], xmm4;  ;// store f[7] in scratch space
// xmm4 f[7] free (in scratch)

    ;

    mulps xmm6, [_PTAN2];       ;// (2a) xmm6 <= (dp[2]*tg_2_16f)

    ;

    mulps xmm7, [_PCOS4]        ;// (4b) xmm7 <= f[6] = (e[5] - e[6]) * cos_4_16f;
    
    ;

    subps xmm6, xmm2;           ;// (2b) xmm6 <= b[3] = (dp[2]*tg_2_16f) - dp[3]
// xmm2 free
// xmm2 used
//movups [SCRATCH+1*16], xmm5;  ;// xmm5 <= store f[5] in scratch space
// xmm5 f[5] free (in scratch)
     movaps xmm2, [INPC+4*32];  ;// (1) xmm2 <= dp[1]

// xmm0 used
    movaps xmm0, xmm1;          ;// (1) xmm0 <= dp[0]
     ;

    addps xmm1, xmm2;           ;// (2b) xmm1 <= b[0] = dp[0] + dp[1]

// movups [SCRATCH+2*16], xmm7; ;// xmm7 <= store f[6] in scratch space
// xmm7 f[6] free (in scratch)

    ;

//    movups [INP+1*16], xmm1;  ;// temp store

    subps xmm0, xmm2;           ;// (2b) xmm0 <= b[1] = dp[0] - dp[1]
// xmm2 free

//    e[0] = b[0] + b[2];
//    e[1] = b[1] + b[3];
//    e[2] = b[1] - b[3];
//    e[3] = b[0] - b[2];

    ;

// xmm2 used 
    movaps xmm2, xmm1;          ;// (2b) xmm1 <= b[0]
     addps xmm1, xmm3;          ;// (3) xmm1 <= e[0] = b[0] + b[2];

    ;

    subps xmm2, xmm3;           ;// (3) xmm3 <= e[3] = b[0] - b[2];
// xmm3 free
// xmm3 used
     movaps xmm3, xmm0;         ;// (2b) xmm3 <= b[1]

    ;

    addps xmm0, xmm6;           ;// (3) xmm0 <= e[1] = b[1] + b[3];

    ;

    subps xmm3, xmm6;           ;// (3) xmm3 <= e[2] = b[1] - b[3];
// xmm6 free
// xmm6 used
     movaps xmm6, xmm2;         ;// (4) xmm6 <= f[3]

// xmm1 <= f[0]
// xmm0 <= f[1]
// xmm3 <= f[2]
// xmm2 <= f[3]
/*

    y[0*8] = SHIFT_ROUND_COLF( _F0 + _F4 );
    y[1*8] = SHIFT_ROUND_COLF( _F1 + _F5 );
    y[2*8] = SHIFT_ROUND_COLF( _F2 + _F6 );
    y[3*8] = SHIFT_ROUND_COLF( _F3 + _F7 );

    y[4*8] = SHIFT_ROUND_COLF( _F3 - _F7 );
    y[5*8] = SHIFT_ROUND_COLF( _F2 - _F6 );
    y[6*8] = SHIFT_ROUND_COLF( _F1 - _F5 );
    y[7*8] = SHIFT_ROUND_COLF( _F0 - _F4 );

    6a) round compensation, adjust float value to correct round alignment
        addps (ROUND_COMP - 0.5)

    6b) float -> int32 conversion (cvtps2pi)
       cvtps2pi mm0, xmm0;  // mm0 <= int32 versions of y[1 0]
        movhlps xmm0, xmm0; 
       cvtps2pi mm1, xmm0;  // mm1 <= int32 versions of y[3 2]

    6c) int32 -> int16 conversion (packssdw)
        packssdw mm0, mm1; // mm0 <= int16 versions of y[3 2 1 0]

    6d) int16 decsale (psraw)
       psraw mm0, DESCALE_SHIFT2; // mm0 <= final descaled output, {-256,+255}

    ** The descale shift is chosen to implement a {-256,+255} output range.

*/
    ;

    addps xmm2, xmm4;           ;// (5) xmm2 <= y[3] = f[3] + f[7]

    ;

    subps xmm6, xmm4;           ;// (5) xmm6 <= y[4] = f[3] - f[7]
// xmm4 free
// xmm4 used
     movaps xmm4, xmm0;         ;// xmm0 <= f[1]

    ;

    addps xmm2, [_PRND_INV_ROW]; // (6a) ROUND compensation

    ;

    addps xmm0, xmm5;           ;// (5) xmm0 <= y[1] = f[1] + f[5]

    ;

   cvtps2pi mm0, xmm2;          ;// (6b) mm0 <= y[3] low
    movhlps xmm2, xmm2;

   cvtps2pi mm1, xmm2;          ;// (6b) mm1 <= y[3] high
    ;

// xmm2 free
   addps xmm0, [_PRND_INV_ROW]; ;// (6a) ROUND compensation

   ;

   subps xmm4, xmm5;            ;// (5) xmm4 <= y[6] = f[1] - f[5]
// xmm5 free
// xmm2 used
    movaps xmm2, [SCRATCH];     ;// xmm2 <= f[4]

   packssdw mm0, mm1;           ;// (6c) mm0 <= y[3] 
    ;

   cvtps2pi mm2, xmm0;          ;// (6b) mm2 <= y[1] low
    movhlps xmm0, xmm0;

   cvtps2pi mm3, xmm0;          ;// (6b) mm3 <= y[1] high
// xmm5 used
    movaps xmm5, xmm3;          ;// xmm5 <= f[2]

   addps xmm3, xmm7;            ;// (5) xmm3 <= y[2] = f[2] + f[6]
    ;

   psraw mm0, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   addps xmm6, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    ;

   packssdw mm2, mm3;           ;// (6c) mm2 <= y[1]
    ;

   subps xmm5, xmm7;            ;// (5) xmm5 <= y[5] = f[2] - f[6]
// xmm7 free
    movq [OUTC + 3*16], mm0;    ;// mm0 <= y[3] stored

   psraw mm2, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   cvtps2pi mm4, xmm6;          ;// (6b) mm4 <= y[4] low
    movhlps xmm6, xmm6;

   cvtps2pi mm5, xmm6;          ;// (6b) mm5 <= y[4] high
// xmm6 free
    ;
  
   addps xmm3, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    movq [OUTC + 1*16], mm2;    ;// y[1] stored

// xmm7 used
   movaps xmm7, xmm1;           ;// xmm7 <= f[0]
    ;

   addps xmm1, xmm2;            ;// (5) xmm2 <= y[0] = f[0] + f[4]
    ;

   packssdw mm4, mm5;           ;// (6c) mm4 <= y[4]
    ;

   cvtps2pi mm6, xmm3;          ;// (6b) mm6 <= y[2] low
    movhlps xmm3, xmm3;

   cvtps2pi mm7, xmm3;          ;// (6b) mm7 <= y[2] high
// xmm3 free
    ;

   subps xmm7, xmm2;            ;// (5) xmm7 <= y[7] = f[0] - f[4]
// xmm2 free

   psraw mm4, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   addps xmm5, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    movq [OUTC + 4*16], mm4;    ;// y[4] stored

   packssdw mm6, mm7;           ;// (6c) mm6 <= y[2]
    ;

   addps xmm4, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    add INPC, 16;

   psraw mm6, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   cvtps2pi mm0, xmm5;          ;// (6b) mm0 <= y[5] low
    movhlps xmm5, xmm5;

   cvtps2pi mm1, xmm5;          ;// (6b) mm1 <= y[5] high
// xmm5 free
    movq [OUTC + 2*16], mm6;    ;// y[2] stored

   cvtps2pi mm2, xmm4;          ;// (6b) mm2 <= y[6] low
    movhlps xmm4, xmm4;

   cvtps2pi mm3, xmm4;          ;// (6b) mm3 <= y[6] high
// xmm4 free
    ;

   addps xmm1, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    ;

   packssdw mm0, mm1;           ;// (6c) mm0 <= y[5]
    ;

   addps xmm7, [_PRND_INV_ROW]; ;// (6a) ROUND compensation
    ;

   psraw mm0, DESCALE_SHIFT2;    // (6d) clip y[] to {-256,+255}
    ;

   cvtps2pi mm4, xmm1;          ;// (6b) mm4 <= y[0] low
    movhlps xmm1, xmm1;

   packssdw mm2, mm3;           ;// (6c) mm2 <= y[6]
    ;

   cvtps2pi mm5, xmm1;          ;// (6b) mm5 <= y[0] high
// xmm1 free
    movq [OUTC + 5*16], mm0;

   psraw mm2, DESCALE_SHIFT2;   // (6d) clip y[] to {-256,+255}
    ;

   cvtps2pi mm6, xmm7;         ;// (6b) mm6 <= y[7] low
    movhlps xmm7, xmm7;

   cvtps2pi mm7, xmm7;         ;// (6b) mm7 <= y[7] high
// xmm7 free
    ;

   packssdw mm4, mm5;          ;// (6c) mm4 <= y[0]
    ;

   psraw mm4, DESCALE_SHIFT2;   // (6d) clip y[] to {-256,+255}
    movq [OUTC + 6*16], mm2;

   packssdw mm6, mm7;          ;// (6c) mm6 <= y[7]
    ;

   psraw mm6, DESCALE_SHIFT2;   // (6d) clip y[] to {-256,+255}
    movq [OUTC + 0*16], mm4;

   movq [OUTC + 7*16], mm6;

 //   ldmxcsr [_mxcsr_backup]; // restore state

   emms;
   }
}